import babelGenerator from "@babel/generator"
import { parse } from "@babel/parser"
import babelTraverse from "@babel/traverse"
import t from "@babel/types"
import { assert, ensure } from "@samual/lib/assert"
import { spliceString } from "@samual/lib/spliceString"
import { tokenizer, tokTypes } from "acorn"
import { resolve } from "import-meta-resolve"
import { validDBMethods } from "../constants.js"
const { default: traverse } = babelTraverse,
	{ default: generate } = babelGenerator,
	SUBSCRIPT_PREFIXES = ["s", "fs", "4s", "hs", "3s", "ms", "2s", "ls", "1s", "ns", "0s"],
	PREPROCESSOR_NAMES = [...SUBSCRIPT_PREFIXES, "D", "G", "FMCL", "db"]
async function preprocess(code, { uniqueId = "00000000000" } = {}) {
	assert(/^\w{11}$/.test(uniqueId), "src/processScript/preprocess.ts:28:36")
	const sourceCode = code,
		tokens = [...tokenizer(code, { ecmaVersion: "latest" })],
		needExportDefault =
			ensure(tokens[0], "src/processScript/preprocess.ts:34:21").type == tokTypes._function &&
			ensure(tokens[1], "src/processScript/preprocess.ts:34:77").type == tokTypes.parenL,
		maybePrivatePrefix = `$${uniqueId}$MAYBE_PRIVATE$`
	for (const token of [...tokens].reverse()) {
		assert("value" in token, "src/processScript/preprocess.ts:39:28")
		if (token.type == tokTypes.privateId) {
			assert("string" == typeof token.value, "src/processScript/preprocess.ts:42:43")
			PREPROCESSOR_NAMES.includes(token.value) &&
				(code = spliceString(code, maybePrivatePrefix + token.value, token.start, token.end - token.start))
		}
	}
	needExportDefault && (code = "export default " + code)
	let program
	traverse(
		parse(code, {
			plugins: [
				"typescript",
				["decorators", { decoratorsBeforeExport: !0 }],
				"doExpressions",
				"functionBind",
				"functionSent",
				"partialApplication",
				["pipelineOperator", { proposal: "hack", topicToken: "%" }],
				"throwExpressions",
				["recordAndTuple", { syntaxType: "hash" }],
				"classProperties",
				"classPrivateProperties",
				"classPrivateMethods",
				"logicalAssignment",
				"numericSeparator",
				"nullishCoalescingOperator",
				"optionalChaining",
				"optionalCatchBinding",
				"objectRestSpread"
			],
			sourceType: "module"
		}),
		{
			Program(path) {
				program = path
			},
			Identifier(path) {
				if (!path.node.name.startsWith(maybePrivatePrefix)) return
				const name = path.node.name.slice(maybePrivatePrefix.length)
				if ("ClassProperty" == path.parent.type && path.parent.key == path.node)
					path.parentPath.replaceWith(
						t.classPrivateProperty(
							t.privateName(t.identifier(name)),
							path.parent.value,
							path.parent.decorators,
							path.parent.static
						)
					)
				else if ("MemberExpression" == path.parent.type)
					if (path.parent.property == path.node) {
						assert(!path.parent.computed, "src/processScript/preprocess.ts:95:36")
						path.replaceWith(t.privateName(t.identifier(name)))
					} else {
						assert(path.parent.object == path.node, "src/processScript/preprocess.ts:98:46")
						if ("db" == name) {
							if (path.parent.computed)
								throw Error(
									"Index notation cannot be used on #db, must be in the form of #db.<DB method name>"
								)
							if ("Identifier" != path.parent.property.type)
								throw Error(
									`Expected DB method name to be an Identifier, got ${path.parent.property.type} instead`
								)
							if (!validDBMethods.includes(path.parent.property.name))
								throw Error("Invalid DB method #db." + path.parent.property.name)
							path.node.name = "$db"
						} else {
							assert(SUBSCRIPT_PREFIXES.includes(name), "src/processScript/preprocess.ts:112:49")
							if (path.parent.computed)
								throw Error(
									`Index notation cannot be used for subscripts, must be in the form of #${name}.foo.bar`
								)
							if ("Identifier" != path.parent.property.type)
								throw Error(
									`Expected subscript user name to be Identifier but got ${path.parent.property.type} instead`
								)
							if ("MemberExpression" != path.parentPath.parent.type)
								throw Error(`Subscripts must be in the form of #${name}.foo.bar`)
							if (path.parentPath.parent.computed)
								throw Error(
									`Index notation cannot be used for subscripts, must be in the form of #${name}.foo.bar`
								)
							if ("Identifier" != path.parentPath.parent.property.type)
								throw Error(
									`Expected subscript subname to be Identifier but got ${path.parent.property.type} instead`
								)
							"CallExpression" == path.parentPath.parentPath?.parent.type &&
							"scripts" == path.parent.property.name &&
							"quine" == path.parentPath.parent.property.name
								? ensure(
										path.parentPath.parentPath.parentPath,
										"src/processScript/preprocess.ts:134:54"
									).replaceWith(t.stringLiteral(sourceCode))
								: (path.node.name = "$" + name)
						}
					}
				else if (
					"BinaryExpression" == path.parent.type &&
					path.parent.left == path.node &&
					"in" == path.parent.operator
				)
					path.replaceWith(t.privateName(t.identifier(name)))
				else if ("ClassMethod" == path.parent.type && path.parent.key == path.node) {
					assert("constructor" != path.parent.kind, "src/processScript/preprocess.ts:142:47")
					path.parentPath.replaceWith(
						t.classPrivateMethod(
							path.parent.kind,
							t.privateName(t.identifier(name)),
							path.parent.params,
							path.parent.body,
							path.parent.static
						)
					)
				} else if ("FMCL" == name) path.node.name = `$${uniqueId}$FMCL$`
				else if ("G" == name) path.node.name = `$${uniqueId}$GLOBAL$`
				else {
					if ("D" != name) {
						if ("db" == name) throw Error("Invalid #db syntax, must be in the form of #db.<DB method name>")
						assert(SUBSCRIPT_PREFIXES.includes(name), "src/processScript/preprocess.ts:161:51 " + name)
						throw Error(`Invalid subscript syntax, must be in the form of #${name}.foo.bar`)
					}
					path.node.name = "$D"
				}
			}
		}
	)
	const needRecord = program.scope.hasGlobal("Record"),
		needTuple = program.scope.hasGlobal("Tuple")
	;(needRecord || needTuple) &&
		program.node.body.unshift(
			t.importDeclaration(
				needRecord
					? needTuple
						? [
								t.importSpecifier(t.identifier("Record"), t.identifier("Record")),
								t.importSpecifier(t.identifier("Tuple"), t.identifier("Tuple"))
							]
						: [t.importSpecifier(t.identifier("Record"), t.identifier("Record"))]
					: [t.importSpecifier(t.identifier("Tuple"), t.identifier("Tuple"))],
				t.stringLiteral("@bloomberg/record-tuple-polyfill")
			)
		)
	program.scope.hasGlobal("Proxy") &&
		program.node.body.unshift(
			t.importDeclaration(
				[t.importDefaultSpecifier(t.identifier("Proxy"))],
				t.stringLiteral(resolve("proxy-polyfill/src/proxy.js", import.meta.url).slice(7))
			)
		)
	if (1 == program.node.body.length && "FunctionDeclaration" == program.node.body[0].type)
		throw Error(
			"Scripts that only contain a single function declaration are no longer supported.\nPrefix the function declaration with `export default`."
		)
	return { code: generate(program.node).code }
}
export { preprocess }
